import torch
import torchvision.transforms as transforms
from PIL import Image
import torch.nn as nn
import torchvision.models as models


class GrayScaleResNet(nn.Module):
    def __init__(self):
        super(GrayScaleResNet, self).__init__()
        self.resnet = models.resnet18(pretrained=False)
        self.resnet.conv1 = nn.Conv2d(
            1, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False
        )
        self.num_classes = 7
        num_ftrs = self.resnet.fc.in_features
        self.resnet.fc = nn.Linear(num_ftrs, self.num_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.resnet(x)


class ImgClassificationTension:
    def __init__(self, model_path: str, num_candidates: int = 2):
        self.num_classes = 7
        self.num_candidates = num_candidates
        self.model = self.load_model(model_path)
        self.transform = self.setup_transform()

    def setup_transform(self) -> transforms.Compose:
        """
        画像の前処理設定
        """
        transform = transforms.Compose(
            [
                transforms.Resize([224, 224]),
                transforms.Grayscale(num_output_channels=1),
                transforms.ToTensor(),
                transforms.Normalize(mean=[0.485], std=[0.229]),
            ]
        )
        return transform

    def load_model(self, model_path: str):
        """
        モデルの読み込みと初期化
        """
        model = GrayScaleResNet()
        model_dict = model.state_dict()
        pretrained_dict = torch.load(model_path)
        pretrained_dict = {
            k: v
            for k, v in pretrained_dict.items()
            if k in model_dict and v.size() == model_dict[k].size()
        }
        model_dict.update(pretrained_dict)
        model.load_state_dict(model_dict, strict=False)
        model.eval()
        return model

    def predict(self, image_path):
        """
        画像を予測する関数
        """
        image = Image.open(image_path).convert("L")
        tensor_image = self.transform(image).unsqueeze(0)
        outputs = self.model(tensor_image)
        probabilities = torch.nn.functional.softmax(outputs, dim=1)
        top_prob, top_classes = torch.topk(probabilities, self.num_candidates)
        top_prob_percent = [round(prob.item() * 100, 2) for prob in top_prob[0]]
        class_names = ["5", "4", "3", "2", "1", "0", "6"]
        predictions = [
            (class_names[class_idx], prob)
            for class_idx, prob in zip(top_classes[0], top_prob_percent)
        ]
        results = []
        for class_name, prob in predictions:
            results.append(f"{class_name}: {prob}%")
        print(results)
        return results


# # 使用例
# model_path = "ResNet152_weights.pth"  # モデルのパス
# img_path = "path/to/image.jpg"  # 画像のパス
# classifier = ImgClassificationTension(model_path)
# predictions = classifier.predict(img_path)
# for prediction in predictions:
#     print(prediction)



from img_classification import ImgClassificationTension

# インスタンスの作成と画像の推論
classifier = ImgClassificationTension("ResNet152_weights.pth", 2)
results = classifier.predict("../IMG_3824.JPG")
print(results)



# PyPIの設定ファイル
# setup.pyを書く必要があるのは、importする際の環境構築を楽にするため + そのコードを用意するのが容易になるから
# 作り方はこちらのURLを見る
## https://qiita.com/Tadahiro_Yamamura/items/2cbcd272a96bb3761cc8

from setuptools import setup

setup(
    name="classification_tension_for_image",
    version="0.0.3",
    install_requires=["packageA", "packageB"],
    description="""
    ユーザの顔写真からユーザのテンションを7段階に分類するモデル
    Model that classifies user tension into seven levels based on their facial photo

    """,
    long_description="""
    このモデルは、ユーザーの顔画像からユーザのテンションを7段階に分類する。このモデルは、ResNet152をベースにしたモデルで、学習済みの重みを用いている。このモデルは、PyTorchを用いて実装されている。
    入力：画像のパス
    出力：ユーザのテンションを7段階に分類した結果の上位2つのクラスとその確率(これについて変更したい場合は、img_classificationのnum_candidatesを変更してください)

    モデルの詳細については以下のリンクを参照してください。(日本語です。)
    https://docs.google.com/presentation/d/1GaxtnU9zE2ymMLbLYAQpq1QF3LJJaoo1cY_HZUNi_vc/edit?usp=sharing 

    This model classifies user tension into seven levels based on their facial images. It is based on the ResNet152 model and utilizes pretrained weights. The model is implemented using PyTorch.
    Input: Path to the image
    Output: Top two classes of user tension levels and their probabilities (to modify this, please change the num_candidates in img_classification)

    For more details about the model, please refer to the following link (content is in Japanese).
    https://docs.google.com/presentation/d/1GaxtnU9zE2ymMLbLYAQpq1QF3LJJaoo1cY_HZUNi_vc/edit?usp=sharing

    """,
    keywords="feeling, person face images, classification",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Communications :: Chat",
    ],
)

# install_requires: pip install -eで一緒にinstallされる
# entry_points: pip install したときに実行可能ファイルとして生成される
